
#include <cstdio>
#include <iostream>
#include <cstring>

#include <id3/tag.h>
#include <id3/misc_support.h>

#include <netinet/in.h>

#include <ruby.h>

typedef struct tag_value {
	ID3_Tag * t ;
	VALUE frames ;
} TagValue ;

typedef struct frame_value {
	VALUE obj ;
	int valid_p ;
	int child_p ;
	TagValue * parent ;
	ID3_Frame * f ;
	VALUE fields ;
} FrameValue ;

typedef struct field_value {
	VALUE obj ;
	int valid_p ;
	FrameValue * parent ;
	ID3_Field * f ;
} FieldValue ;


VALUE rb_mID3Lib ;

VALUE rb_cID3Tag ;
VALUE rb_cID3Frame ;
VALUE rb_cID3Field ;

static void id3lib_ID3Tag_remove_child_frame(TagValue * tv, FrameValue * fv) ;
static void id3lib_ID3Frame_remove_child_field(FrameValue * fv, FieldValue * flv) ;

static bool
valid_FrameID_p(ID3_FrameID fid)
{
	switch(fid) {
	case ID3FID_NOFRAME:        case ID3FID_AUDIOCRYPTO:   
	case ID3FID_PICTURE:        case ID3FID_AUDIOSEEKPOINT:
	case ID3FID_COMMENT:        case ID3FID_COMMERCIAL:    
	case ID3FID_CRYPTOREG:      case ID3FID_EQUALIZATION2: 
	case ID3FID_EQUALIZATION:   case ID3FID_EVENTTIMING:   
	case ID3FID_GENERALOBJECT:  case ID3FID_GROUPINGREG:   
	case ID3FID_INVOLVEDPEOPLE: case ID3FID_LINKEDINFO:    
	case ID3FID_CDID:           case ID3FID_MPEGLOOKUP:    
	case ID3FID_OWNERSHIP:      case ID3FID_PRIVATE:       
	case ID3FID_PLAYCOUNTER:    case ID3FID_POPULARIMETER: 
	case ID3FID_POSITIONSYNC:   case ID3FID_BUFFERSIZE:    
	case ID3FID_VOLUMEADJ2:     case ID3FID_VOLUMEADJ:     
	case ID3FID_REVERB:         case ID3FID_SEEKFRAME:     
	case ID3FID_SIGNATURE:      case ID3FID_SYNCEDLYRICS:  
	case ID3FID_SYNCEDTEMPO:    case ID3FID_ALBUM:         
	case ID3FID_BPM:            case ID3FID_COMPOSER:      
	case ID3FID_CONTENTTYPE:    case ID3FID_COPYRIGHT:     
	case ID3FID_DATE:           case ID3FID_ENCODINGTIME:  
	case ID3FID_PLAYLISTDELAY:  case ID3FID_ORIGRELEASETIME: 
	case ID3FID_RECORDINGTIME:  case ID3FID_RELEASETIME:     
	case ID3FID_TAGGINGTIME:    case ID3FID_INVOLVEDPEOPLE2: 
	case ID3FID_ENCODEDBY:      case ID3FID_LYRICIST:        
	case ID3FID_FILETYPE:       case ID3FID_TIME:            
	case ID3FID_CONTENTGROUP:   case ID3FID_TITLE:           
	case ID3FID_SUBTITLE:       case ID3FID_INITIALKEY:      
	case ID3FID_LANGUAGE:       case ID3FID_SONGLEN:         
	case ID3FID_MUSICIANCREDITLIST:
	case ID3FID_MEDIATYPE:      case ID3FID_MOOD:              
	case ID3FID_ORIGALBUM:      case ID3FID_ORIGFILENAME:      
	case ID3FID_ORIGLYRICIST:   case ID3FID_ORIGARTIST:        
	case ID3FID_ORIGYEAR:       case ID3FID_FILEOWNER:         
	case ID3FID_LEADARTIST:     case ID3FID_BAND:              
	case ID3FID_CONDUCTOR:      case ID3FID_MIXARTIST:         
	case ID3FID_PARTINSET:      case ID3FID_PRODUCEDNOTICE:    
	case ID3FID_PUBLISHER:      case ID3FID_TRACKNUM:          
	case ID3FID_RECORDINGDATES: case ID3FID_NETRADIOSTATION:   
	case ID3FID_NETRADIOOWNER:  case ID3FID_SIZE:              
	case ID3FID_ALBUMSORTORDER: case ID3FID_PERFORMERSORTORDER:
	case ID3FID_TITLESORTORDER: case ID3FID_ISRC:              
	case ID3FID_ENCODERSETTINGS:   
	case ID3FID_USERTEXT:          
	case ID3FID_YEAR:           case ID3FID_UNIQUEFILEID:      
	case ID3FID_TERMSOFUSE:     case ID3FID_UNSYNCEDLYRICS:    
	case ID3FID_WWWCOMMERCIALINFO: 
	case ID3FID_WWWCOPYRIGHT:   case ID3FID_WWWAUDIOFILE:      
	case ID3FID_WWWARTIST:      case ID3FID_WWWAUDIOSOURCE:    
	case ID3FID_WWWRADIOPAGE:   case ID3FID_WWWPAYMENT:        
	case ID3FID_WWWPUBLISHER:   case ID3FID_WWWUSER:           
	case ID3FID_METACRYPTO:     case ID3FID_METACOMPRESSION:   
	case ID3FID_LASTFRAMEID:
		return true ;
	default:
		return false ;
	}
	
	return false ;
}

static bool
valid_FieldID_p(ID3_FieldID fid)
{
	switch(fid) {
	case ID3FN_NOFIELD:
	case ID3FN_TEXTENC:
	case ID3FN_TEXT:
	case ID3FN_URL:
	case ID3FN_DATA:
	case ID3FN_DESCRIPTION:
	case ID3FN_OWNER:
	case ID3FN_EMAIL:
	case ID3FN_RATING:
	case ID3FN_FILENAME:
	case ID3FN_LANGUAGE:
	case ID3FN_PICTURETYPE:
	case ID3FN_IMAGEFORMAT:
	case ID3FN_MIMETYPE:
	case ID3FN_COUNTER:
	case ID3FN_ID:
	case ID3FN_VOLUMEADJ:
	case ID3FN_NUMBITS:
	case ID3FN_VOLCHGRIGHT:
	case ID3FN_VOLCHGLEFT:
	case ID3FN_PEAKVOLRIGHT:
	case ID3FN_PEAKVOLLEFT:
	case ID3FN_TIMESTAMPFORMAT:
	case ID3FN_CONTENTTYPE:
	case ID3FN_LASTFIELDID:
		return true ;
	default:
		return false ;
	}
	
	return false ;
}


static void
id3lib_ID3Field_invalidate(FieldValue * fv)
{
	fv->valid_p = 0 ;
	fv->f = 0 ;
}

static void
id3lib_ID3Field_delete(FieldValue * fv)
{
	if (!fv->valid_p) {
		free(fv) ;
		return  ;
	}
	
	id3lib_ID3Frame_remove_child_field(fv->parent, fv) ;
	free(fv) ;
	return ;
}

static FieldValue *
getID3FieldValue(VALUE v)
{
	FieldValue * fv ;
	
	switch(TYPE(v)) {
	case T_DATA:
		if (RDATA(v)->dfree == (void *)id3lib_ID3Field_delete) {
			Data_Get_Struct(v, FieldValue, fv) ;
			return fv ;
		}
	default:
		rb_bug("It isn't an ID3Field Object...") ;
	}
	
	return NULL ;
}

static VALUE
id3lib_ID3Field_new_object(VALUE klass, ID3_Field* field, FrameValue * parent)
{
	VALUE v ;
	FieldValue * fv ;
	
	fv = ALLOC(FieldValue) ;
	fv->valid_p = 1 ;
	fv->parent = parent ;
	fv->f = field ;
	
	v = (VALUE)Data_Wrap_Struct(klass, 0, id3lib_ID3Field_delete, fv) ;
	fv->obj = v ;
	return v ;
}

static VALUE
id3lib_ID3Field_clear(VALUE self)
{
	FieldValue * fv = getID3FieldValue(self) ;
	
	if (fv->valid_p) {
		fv->f->Clear( ) ;
	}
	
	return Qnil ;
}


static char *
unicode_to_char(const unicode_t * ustr, size_t len)
{
	int ulen = len>>1 ;
	char * str = new char[len+2] ;
	
	for (int i=0; i<ulen; i++) {
		((uint16_t *)str)[i] = htons((uint16_t)ustr[i]) ;
	}
	((uint16_t *)str)[ulen] = 0 ;
	return str ;
}


static unicode_t *
char_to_unicode(const char * str, size_t len)
{
	int ulen = len>>1 ;
	unicode_t * ustr = new unicode_t[ulen+1] ;
	
	for (int i=0; i<ulen; i++) {
		ustr[i] = (unicode_t)ntohs(*(uint16_t *)(str+i*2)) ;
	}
	ustr[ulen] = 0 ;
	return ustr ;
}


static VALUE
id3lib_ID3Field_get(VALUE self)
{
	FieldValue * fv = getID3FieldValue(self) ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	ID3_FieldType ft = fv->f->GetType( ) ;
	switch (ft) {
	case ID3FTY_NONE:
		return Qnil ;
		
	case ID3FTY_INTEGER:
	{
		uint32 x = fv->f->Get( ) ;
		return INT2NUM(x) ;
	}
	
	case ID3FTY_BINARY:
	{
		const uchar * bin = fv->f->GetRawBinary( ) ;
		size_t len = fv->f->BinSize( ) ;
		return rb_str_new((const char *)bin, len) ;
	}
	
	case ID3FTY_TEXTSTRING:
	{
		if (fv->f->IsEncodable( )) {
			size_t len = fv->f->Size( ) ;
			
			switch (fv->f->GetEncoding( )) {
			case ID3TE_ASCII:
				return rb_str_new(fv->f->GetRawText( ), len) ;
				
			case ID3TE_UNICODE:
			{
				char * str = unicode_to_char(fv->f->GetRawUnicodeText( ), len) ;
				volatile VALUE v = rb_str_new(str, len) ;
				delete str ;
				return v ;
			}
			
			default:
				return Qnil ;
			}
		} else {
			size_t len = fv->f->Size( ) ;
			return rb_str_new(fv->f->GetRawText( ), len) ;
		}
		
	}
	default:
		rb_raise(rb_eRuntimeError, "ID3Field#get: Unknown field type %d.", ft) ;
	}
	
	return Qnil ;
}


#define FIELD_VALID_CHECK()  { if (!fv->valid_p) { return Qnil ; } }

static VALUE
id3lib_ID3Field_set(VALUE self, VALUE v)
{
	size_t ret = 0 ;
	FieldValue * fv = getID3FieldValue(self) ;
	
	FIELD_VALID_CHECK( ) ;
	
	ID3_FieldType ft = fv->f->GetType( ) ;
	switch (ft) {
	case ID3FTY_NONE:
		break ;
		
	case ID3FTY_INTEGER:
	{
		uint32 x = NUM2UINT(v) ;
		FIELD_VALID_CHECK( ) ;
		fv->f->Set(x) ;
		ret = sizeof(uint32) ;
		break ;
	}
	
	case ID3FTY_BINARY:
	{
		volatile VALUE str = StringValue(v) ;
		uchar * bin = (uchar *)RSTRING(str)->ptr ;
		size_t len = RSTRING(str)->len ;
		FIELD_VALID_CHECK( ) ;
		if (bin != NULL) {
			ret = fv->f->Set(bin, len) ;
		}
		break ;
	}
	
	case ID3FTY_TEXTSTRING:
	{
		FIELD_VALID_CHECK( ) ;
		if (fv->f->IsEncodable( )) {
			volatile VALUE str = StringValue(v) ;
			size_t len = RSTRING(str)->len ;
			FIELD_VALID_CHECK( ) ;
			
			switch (fv->f->GetEncoding( )) {
			case ID3TE_ASCII:
				if (RSTRING(str)->ptr) {
					char * sp = new char[len+1] ;
					memcpy(sp, RSTRING(str)->ptr, len+1) ;
					ret = fv->f->Set(sp) ;
					delete sp ;
				}
				break ;
			case ID3TE_UNICODE:
			{
				if (RSTRING(str)->ptr) {
					unicode_t * ustr = char_to_unicode(RSTRING(str)->ptr, len) ;
					ret = fv->f->Set(ustr) ;
					delete ustr ;
				} else {
					ret = fv->f->Set((unicode_t *)"\0\0") ;
				}
				
				break ;
			}
			default:
				ret = 0 ;
				break ;
			}
			break ;
		} else {
			volatile VALUE str = StringValue(v) ;
			size_t len = RSTRING(str)->len ;
			FIELD_VALID_CHECK( ) ;
			
			if (RSTRING(str)->ptr) {
				char * sp = new char[len+1] ;
				memcpy(sp, RSTRING(str)->ptr, len+1) ;
				ret = fv->f->Set(sp) ;
				delete sp ;
			} else {
				ret = fv->f->Set("") ;
			}
			break ;
		}
	}
		
	default:
		rb_raise(rb_eRuntimeError, "ID3Field#set: Unknown field type %d.", ft) ;
	}
	
	return INT2NUM(ret) ;
}


static VALUE
id3lib_ID3Field_isEncodable_p(VALUE self)
{
	FieldValue * fv = getID3FieldValue(self) ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	if (fv->f->IsEncodable( )) {
		return Qtrue ;
	}
	return Qfalse ;
}

static VALUE
id3lib_ID3Field_getEncoding(VALUE self)
{
	FieldValue * fv = getID3FieldValue(self) ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	return INT2NUM(fv->f->GetEncoding( )) ;
}

static VALUE
id3lib_ID3Field_setEncoding(VALUE self, VALUE enc)
{
	FieldValue * fv = getID3FieldValue(self) ;
	ID3_TextEnc e = (ID3_TextEnc)NUM2INT(enc) ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	if (e != ID3TE_ASCII && e != ID3TE_UNICODE) {
		rb_raise(rb_eRuntimeError, "ID3Field#setEncoding: Unknown text encoding %d.", e) ;
	}
	
	if (fv->f->SetEncoding(e)) {
		return Qtrue ;
	}
	
	return Qfalse ;
}




static void
id3lib_ID3Frame_clear_fields(FrameValue * fv)
{
	VALUE ary = fv->fields ;
	int l ;
	int i ;
	
	if (NIL_P(ary)) {
		return ;
	}
	
	l = RARRAY(ary)->len ;
	
	for (i=0; i<l; i++) {
		VALUE v = rb_ary_pop(ary) ;
		id3lib_ID3Field_invalidate(getID3FieldValue(v)) ;
	}
	
}

static VALUE
id3lib_ID3Frame_find_child_field(FrameValue * fv, ID3_Field * field)
{
	int i, l ;
	VALUE result = Qnil ;
	VALUE ary = fv->fields ;
	
	if (NIL_P(ary)) {
		return Qnil ;
	}
	
	l = RARRAY(ary)->len ;
	for (i=0; i<l; i++) {
		VALUE v = RARRAY(ary)->ptr[i] ;
		FieldValue * flv = getID3FieldValue(v) ;
		if (flv->f == field) {
			result = v ;
			break ;
		}
	}
	
	return result ;
}


static VALUE
id3lib_ID3Frame_make_child_field(FrameValue * fv, ID3_Field * field)
{
	VALUE v ;
	VALUE ary = fv->fields ;
	
	if (NIL_P(ary)) {
		ary = rb_ary_new( ) ;
		fv->fields = ary ;
	}
	
	v = id3lib_ID3Field_new_object(rb_cID3Field, field, fv) ;
	rb_ary_push(ary, v) ;
	return v ;
}

static void
id3lib_ID3Frame_remove_child_field(FrameValue * fv, FieldValue * flv)
{
	VALUE ary = fv->fields ;
	VALUE v ;
	
	if (NIL_P(ary)) {
		return ;
	}
	
	v = rb_ary_delete(ary, flv->obj) ;
	if (!NIL_P(v)) {
		flv->parent = 0 ;
	}
}

static void
id3lib_ID3Frame_invalidate(FrameValue * fv)
{
	id3lib_ID3Frame_clear_fields(fv) ;
	fv->valid_p = 0 ;
	fv->f = 0 ;
}

static void
id3lib_ID3Frame_mark(FrameValue * fv)
{
	if (fv->valid_p) {
		rb_gc_mark(fv->fields) ;
	}
}


static void
id3lib_ID3Frame_delete(FrameValue * fv)
{
	if (!fv->valid_p) {
		free(fv) ;
		return ;
	}
	
	if (fv->child_p) {
		id3lib_ID3Tag_remove_child_frame(fv->parent, fv) ;
		free(fv) ;
		return ;
	}
	
	delete fv->f ;
	free(fv) ;
}

static FrameValue *
getID3FrameValue(VALUE v)
{
	FrameValue * fv ;
	
	switch(TYPE(v)) {
	case T_DATA:
		if (RDATA(v)->dfree == (void *)id3lib_ID3Frame_delete) {
			Data_Get_Struct(v, FrameValue, fv) ;
			return fv ;
		}
	default:
		rb_bug("It isn't an ID3Frame Object...") ;
	}
	
	return NULL ;
}

static VALUE
id3lib_ID3Frame_new_object(VALUE self, ID3_Frame* frame, int child_p, TagValue * parent)
{
	VALUE v ;
	FrameValue * fv ;
	
	fv = ALLOC(FrameValue) ;
	fv->valid_p = 1 ;
	fv->child_p = child_p ;
	fv->parent = parent ;
	fv->f = frame ;
	fv->fields = Qnil ;
	
	v = (VALUE)Data_Wrap_Struct(self, id3lib_ID3Frame_mark, id3lib_ID3Frame_delete, fv) ;
	fv->obj = v ;
	return v ;
}

static VALUE
id3lib_ID3Frame_alloc(VALUE klass)
{
	VALUE v ;
	FrameValue * fv ;
	
	fv = ALLOC(FrameValue) ;
	fv->valid_p = 0 ;
	fv->child_p = 0 ;
	fv->parent = 0 ;
	fv->f = 0 ;
	fv->fields = Qnil ;
	
	v = (VALUE)Data_Wrap_Struct(klass, id3lib_ID3Frame_mark, id3lib_ID3Frame_delete, fv) ;
	fv->obj = v ;
	return v ;
}

static VALUE
id3lib_ID3Frame_initialize(int argc, VALUE * argv, VALUE self)
{
	FrameValue * fv = getID3FrameValue(self) ;
	ID3_Frame* frame ;
	VALUE vfid ;
	
	rb_scan_args(argc, argv, "01", &vfid) ;
	
	if (NIL_P(vfid)) {
		frame = new ID3_Frame ;
	} else {
		ID3_FrameID fid = (ID3_FrameID)NUM2INT(vfid) ;
		
		if (!valid_FrameID_p(fid)) {
			rb_raise(rb_eRuntimeError, "ID3Frame: Invalid Frame ID : %d", fid) ;
		}
		
		frame = new ID3_Frame(fid) ;
	}
	
	if (frame == NULL) {
		rb_raise(rb_eRuntimeError, "ID3Frame: Can't allocate ID3Frame struct.") ;
		return Qnil ;
	}
	
	fv->valid_p = 1 ;
	fv->f = frame ;
	
	return self ;
}

static VALUE
id3lib_ID3Frame_init_copy(VALUE copy, VALUE obj)
{
	if (copy == obj) return copy ;
	
	FrameValue * src = getID3FrameValue(obj) ;
	FrameValue * dst = getID3FrameValue(copy) ;
	
	dst->valid_p = 1 ;
	dst->f = new ID3_Frame(*(src->f)) ;
	dst->child_p = 0 ;
	dst->parent = 0 ;
	
	return copy ;
}


static VALUE
id3lib_ID3Frame_clear(VALUE self)
{
	FrameValue * fv = getID3FrameValue(self) ;
	
	if (fv->valid_p) {
		fv->f->Clear( ) ;
		id3lib_ID3Frame_clear_fields(fv) ;
	}
	
	return Qnil ;
}

static VALUE
id3lib_ID3Frame_setID(VALUE self, VALUE vfid)
{
	FrameValue * fv = getID3FrameValue(self) ;
	ID3_FrameID fid = (ID3_FrameID)NUM2INT(vfid) ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	if (!valid_FrameID_p(fid)) {
		rb_raise(rb_eRuntimeError, "ID3Frame: Invalid Frame ID : %d", fid) ;
	}
	
	id3lib_ID3Frame_clear_fields(fv) ;
	fv->f->SetID(fid) ;
	return Qnil ;
}

static VALUE
id3lib_ID3Frame_getID(VALUE self)
{
	FrameValue * fv = getID3FrameValue(self) ;
	ID3_FrameID fid ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	fid = fv->f->GetID( ) ;
	return INT2NUM(fid) ;
}

static VALUE
id3lib_ID3Frame_numFields(VALUE self)
{
	FrameValue * fv = getID3FrameValue(self) ;
	int n ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	n = fv->f->NumFields( ) ;
	return INT2NUM(n) ;
}

static VALUE
id3lib_ID3Frame_getDescription(VALUE self)
{
	FrameValue * fv = getID3FrameValue(self) ;
	const char * desc ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	desc = fv->f->GetDescription( ) ;
	return rb_str_new2(desc) ;
}

static VALUE
id3lib_ID3Frame_getTextID(VALUE self)
{
	FrameValue * fv = getID3FrameValue(self) ;
	const char * tid ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	tid = fv->f->GetTextID( ) ;
	return rb_str_new2(tid) ;
}

static VALUE
id3lib_ID3Frame_field(VALUE self, VALUE fieldID)
{
	FrameValue * fv = getID3FrameValue(self) ;
	ID3_Field * field ;
	ID3_FieldID fid = (ID3_FieldID)NUM2INT(fieldID) ;
	VALUE v ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	if (!valid_FieldID_p(fid)) {
		rb_raise(rb_eRuntimeError, "ID3Frame: Invalid Field ID : %d", fid) ;
	}
	
	field = &(fv->f->Field(fid)) ;
	if (field == NULL) {
		return Qnil ;
	}
	
	v = id3lib_ID3Frame_find_child_field(fv, field) ;
	if (NIL_P(v)) {
		return id3lib_ID3Frame_make_child_field(fv, field) ;
	}
	
	return v ;
}

static VALUE
id3lib_ID3Frame_contains_p(VALUE self, VALUE fieldID)
{
	FrameValue * fv = getID3FrameValue(self) ;
	ID3_FieldID fid = (ID3_FieldID)NUM2INT(fieldID) ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	if (!valid_FieldID_p(fid)) {
		rb_raise(rb_eRuntimeError, "ID3Frame: Invalid Field ID : %d", fid) ;
	}
	
	if (fv->f->Contains(fid)) {
		return Qtrue ;
	}
	
	return Qfalse ;
}

static VALUE
id3lib_ID3Frame_getCompression(VALUE self)
{
	FrameValue * fv = getID3FrameValue(self) ;
	
	if (!fv->valid_p) {
		return Qnil ;
	}
	
	if (fv->f->GetCompression( )) {
		return Qtrue ;
	}
	
	return Qfalse ;
}

static VALUE
id3lib_ID3Frame_setCompression(VALUE self, VALUE b)
{
	FrameValue * fv = getID3FrameValue(self) ;
	bool p ;
	
	switch (TYPE(b)) {
	case T_TRUE:
		p = true ;
		break ;
		
	case T_FALSE:
	case T_NIL:
		p = false ;
		break ;
		
	default:
		rb_raise(rb_eTypeError,
				 "wrong argument type %s (expected true/false)",
				 rb_class2name(rb_obj_class(b))) ;
	}
	
	if (fv->f->SetCompression(p)) {
		return Qtrue ;
	}
	
	return Qfalse ;
}




static void
id3lib_ID3Tag_clear_frames(TagValue * tv)
{
	VALUE ary = tv->frames ;
	int l ;
	int i ;
	
	if (NIL_P(ary)) {
		return ;
	}
	
	l = RARRAY(ary)->len ;
	
	for (i=0; i<l; i++) {
		VALUE v = rb_ary_pop(ary) ;
		id3lib_ID3Frame_invalidate(getID3FrameValue(v)) ;
	}
	
}

static VALUE
id3lib_ID3Tag_find_child_frame(TagValue * tv, ID3_Frame* frame)
{
	int i, l ;
	VALUE result = Qnil ;
	VALUE ary = tv->frames ;
	
	if (NIL_P(ary)) {
		return Qnil ;
	}
	
	l = RARRAY(ary)->len ;
	for (i=0; i<l; i++) {
		VALUE v = RARRAY(ary)->ptr[i] ;
		FrameValue * fv = getID3FrameValue(v) ;
		if (fv->f == frame) {
			result = v ;
			break ;
		}
	}
	
	return result ;
}

static VALUE
id3lib_ID3Tag_add_child_frame(TagValue * tv, FrameValue * fv)
{
	VALUE ary = tv->frames ;
	
	if (NIL_P(ary)) {
		ary = rb_ary_new( ) ;
		tv->frames = ary ;
	}
	
	fv->child_p = 1 ;
	fv->parent = tv ;
	
	rb_ary_push(ary, fv->obj) ;
	return fv->obj ;
}


static VALUE
id3lib_ID3Tag_make_child_frame(TagValue * tv, ID3_Frame* frame)
{
	VALUE v ;
	VALUE ary = tv->frames ;
	
	if (NIL_P(ary)) {
		ary = rb_ary_new( ) ;
		tv->frames = ary ;
	}
	
	v = id3lib_ID3Frame_new_object(rb_cID3Frame, frame, 1, tv) ;
	rb_ary_push(ary, v) ;
	return v ;
}

static void
id3lib_ID3Tag_remove_child_frame(TagValue * tv, FrameValue * fv)
{
	VALUE ary = tv->frames ;
	VALUE v ;
	
	if (NIL_P(ary)) {
		return ;
	}
	
	v = rb_ary_delete(ary, fv->obj) ;
	if (!NIL_P(v)) {
		fv->child_p = 0 ;
		fv->parent = NULL ;
	}
}

static VALUE
id3lib_ID3Tag_make_frame_array(TagValue * tv)
{
	volatile VALUE ary = rb_ary_new( ) ;
	ID3_Tag::Iterator * iter = tv->t->CreateIterator( ) ;
	
	for (size_t i=0; i<(tv->t->NumFrames( )); i++) {
		ID3_Frame * f = iter->GetNext( ) ;
		if (f == NULL) continue ;
		
		VALUE v = id3lib_ID3Tag_find_child_frame(tv, f) ;
		if (NIL_P(v)) {
			v = id3lib_ID3Tag_make_child_frame(tv, f) ;
		}
		
		rb_ary_push(ary, v) ;
	}
	
	delete iter ;
	return ary ;
}


static void
id3lib_ID3Tag_mark(TagValue * tv)
{
	if (!NIL_P(tv->frames)) {
		rb_gc_mark(tv->frames) ;
	}
}

static void
id3lib_ID3Tag_delete(TagValue * tv)
{
	id3lib_ID3Tag_clear_frames(tv) ;
	delete tv->t ;
	free(tv) ;
}


static TagValue *
getID3TagValue(VALUE v)
{
	TagValue * t ;
	
	switch(TYPE(v)) {
	case T_DATA:
		if (RDATA(v)->dfree == (void *)id3lib_ID3Tag_delete) {
			Data_Get_Struct(v, TagValue, t) ;
			return t ;
		}
	default:
		rb_bug("It isn't an ID3Tag Object...") ;
	}
	
	 return NULL ;
}


static VALUE
id3lib_ID3Tag_alloc(VALUE klass)
{
	VALUE v ;
	TagValue * tv ;
	
	tv = ALLOC(TagValue) ;
	tv->t = 0 ;
	tv->frames = Qnil ;
	v = (VALUE)Data_Wrap_Struct(klass, id3lib_ID3Tag_mark, id3lib_ID3Tag_delete, tv) ;
	return v ;
}

static VALUE
id3lib_ID3Tag_initialize(int argc, VALUE * argv, VALUE self)
{
	VALUE vfname ;
	TagValue * tv = getID3TagValue(self) ;
	
	rb_scan_args(argc, argv, "01", &vfname) ;
	
	ID3_Tag * tag = new ID3_Tag ;
	
	tv->t = tag ;
	
	return self ;
}

static VALUE
id3lib_ID3Tag_init_copy(VALUE copy, VALUE obj)
{
	if (copy == obj) return copy ;
	
	TagValue * src = getID3TagValue(obj) ;
	TagValue * dst = getID3TagValue(copy) ;
	
	dst->t = new ID3_Tag(*(src->t)) ;
	
	return copy ;
}


static VALUE
id3lib_ID3Tag_clear(VALUE self)
{
	TagValue * tag = getID3TagValue(self) ;
	
	tag->t->Clear( ) ;
	
	id3lib_ID3Tag_clear_frames(tag) ;
	
	return Qnil ;
}

static VALUE
id3lib_ID3Tag_strip(int argc, VALUE * argv, VALUE self)
{
	TagValue * tag = getID3TagValue(self) ;
	VALUE vflags ;
	
	rb_scan_args(argc, argv, "01", &vflags) ;
	
	if (NIL_P(vflags)) {
		tag->t->Strip( ) ;
	} else {
		flags_t flags = NUM2INT(vflags) & 0xffff ;
		tag->t->Strip(flags) ;
	}
	
	return Qnil ;
}

static VALUE
id3lib_ID3Tag_update(int argc, VALUE * argv, VALUE self)
{
	TagValue * tag = getID3TagValue(self) ;
	VALUE vflags ;
	
	rb_scan_args(argc, argv, "01", &vflags) ;
	
	if (NIL_P(vflags)) {
		tag->t->Update( ) ;
	} else {
		flags_t flags = NUM2INT(vflags) & 0xffff ;
		tag->t->Update(flags) ;
	}
	
	return Qnil ;
}

static VALUE
id3lib_ID3Tag_link(int argc, VALUE * argv, VALUE self)
{
	TagValue * tag = getID3TagValue(self) ;
	VALUE vfname ;
	VALUE vflags ;
	size_t ret ;
	char * fname ;
	
	rb_scan_args(argc, argv, "11", &vfname, &vflags) ;
	fname = STR2CSTR(vfname) ;
	
	if (NIL_P(vflags)) {
		ret = tag->t->Link(fname) ;
	} else {
		flags_t flags = NUM2INT(vflags) & 0x0ffff ;
		ret = tag->t->Link(fname, flags) ;
	}
	
	return INT2NUM(ret) ;
}

static VALUE
id3lib_ID3Tag_setPadding(VALUE self, VALUE pad)
{
	TagValue * tv = getID3TagValue(self) ;
	bool p ;
	
	switch (TYPE(pad)) {
	case T_TRUE:
		p = true ;
		break ;
		
	case T_FALSE:
	case T_NIL:
		p = false ;
		break ;
		
	default:
		rb_raise(rb_eTypeError,
				 "wrong argument type %s (expected true/false)",
				 rb_class2name(rb_obj_class(pad))) ;
	}
	
	tv->t->SetPadding(p) ;
	return Qnil ;
}

static VALUE
id3lib_ID3Tag_hasChanged_p(VALUE self)
{
	TagValue * tv = getID3TagValue(self) ;
	bool p ;
	
	p = tv->t->HasChanged( ) ;
	if (p) {
		return Qtrue ;
	}
	
	return Qfalse ;
}

static VALUE
id3lib_ID3Tag_hasV1Tag_p(VALUE self)
{
	TagValue * tv = getID3TagValue(self) ;
	bool p = tv->t->HasV1Tag( ) ;
	if (p) {
		return Qtrue ;
	}
	
	return Qfalse ;
}

static VALUE
id3lib_ID3Tag_hasV2Tag_p(VALUE self)
{
	TagValue * tv = getID3TagValue(self) ;
	bool p = tv->t->HasV2Tag( ) ;
	if (p) {
		return Qtrue ;
	}
	
	return Qfalse ;
}

static VALUE
id3lib_ID3Tag_numFrames(VALUE self)
{
	TagValue * tv = getID3TagValue(self) ;
	int n ;
	
	n = tv->t->NumFrames( ) ;
	
	return INT2NUM(n) ;
}


static VALUE
id3lib_ID3Tag_find(int argc, VALUE * argv, VALUE self)
{
	TagValue * tv = getID3TagValue(self) ;
	ID3_Frame* frame ;
	VALUE v ;
	ID3_FrameID frameID ;
	
	rb_scan_args(argc, argv, "10", &v) ;
	frameID = (ID3_FrameID)NUM2INT(v) ;
	
	if (!valid_FrameID_p(frameID)) {
		rb_raise(rb_eRuntimeError, "ID3Frame: Invalid Frame ID : %d", frameID) ;
	}
	
	frame = tv->t->Find(frameID) ;
	
	if (frame == NULL) {
		return Qnil ;
	}
	
	v = id3lib_ID3Tag_find_child_frame(tv, frame) ;
	
	if (NIL_P(v)) {
		return id3lib_ID3Tag_make_child_frame(tv, frame) ;
	}
	
	return v ;
}

static VALUE
id3lib_ID3Tag_find_all(int argc, VALUE * argv, VALUE self)
{
	TagValue * tv = getID3TagValue(self) ;
	VALUE v ;
	VALUE ary ;
	ID3_FrameID frameID ;
	ID3_Tag::Iterator * ite ;
	
	rb_scan_args(argc, argv, "10", &v) ;
	frameID = (ID3_FrameID)NUM2INT(v) ;
	
	if (!valid_FrameID_p(frameID)) {
		rb_raise(rb_eRuntimeError, "ID3Frame: Invalid Frame ID : %d", frameID) ;
	}
	
	ary = rb_ary_new( ) ;
	
	ite = tv->t->CreateIterator( ) ;
	
	while (1) {
		ID3_Frame * frame = ite->GetNext( ) ;
		if (frame == NULL) break ;
		
		if (frame->GetID( ) != frameID) continue ;
		
		v = id3lib_ID3Tag_find_child_frame(tv, frame) ;
		
		if (NIL_P(v)) {
			rb_ary_push(ary, id3lib_ID3Tag_make_child_frame(tv, frame)) ;
		} else {
			rb_ary_push(ary, v) ;
		}
	}
	
	delete ite ;
	
	return ary ;
}

static VALUE
id3lib_ID3Tag_attachFrame(VALUE self, VALUE vframe)
{
	TagValue * tv = getID3TagValue(self) ;
	FrameValue * fv ;
	
	Check_Type(vframe, T_DATA) ;
	if (RDATA(vframe)->dfree != (void *)id3lib_ID3Frame_delete) {
		rb_raise(rb_eTypeError, "ID3Frame object required.") ;
	}
	
	fv = getID3FrameValue(vframe) ;
	
	if (fv->child_p) {
		rb_raise(rb_eRuntimeError, "Already attached frame.") ;
	}
	
	id3lib_ID3Tag_add_child_frame(tv, fv) ;
	tv->t->AttachFrame(fv->f) ;
	
	return self ;
}

static VALUE
id3lib_ID3Tag_addFrame(VALUE self, VALUE vframe)
{
	TagValue * tv = getID3TagValue(self) ;
	FrameValue * fv ;
	
	Check_Type(vframe, T_DATA) ;
	if (RDATA(vframe)->dfree != (void *)id3lib_ID3Frame_delete) {
		rb_raise(rb_eTypeError, "ID3Frame object required.") ;
	}
	
	fv = getID3FrameValue(vframe) ;
	
	id3lib_ID3Tag_add_child_frame(tv, fv) ;
	tv->t->AddFrame(fv->f) ;
	
	return self ;
}


static VALUE
id3lib_ID3Tag_removeFrame(VALUE self, VALUE vframe)
{
	ID3_Frame* ret ;
	TagValue * tv = getID3TagValue(self) ;
	FrameValue * fv ;
	
	Check_Type(vframe, T_DATA) ;
	if (RDATA(vframe)->dfree != (void *)id3lib_ID3Frame_delete) {
		rb_raise(rb_eTypeError, "ID3Frame object required.") ;
	}
	
	fv = getID3FrameValue(vframe) ;
	ret = tv->t->RemoveFrame(fv->f) ;
	if (ret == NULL) {
		return Qnil ;
	}
	
	id3lib_ID3Tag_remove_child_frame(tv, fv) ;
	return fv->obj ;
}


static VALUE
id3lib_ID3Tag_each_frame(VALUE self)
{
	TagValue * tv = getID3TagValue(self) ;
	volatile VALUE v = Qnil ;
	volatile VALUE ary ;
	int i ;
	
	ary = id3lib_ID3Tag_make_frame_array(tv) ;
	
	for (i=0; i < RARRAY(ary)->len; i++) {
		v = rb_yield(RARRAY(ary)->ptr[i]) ;
	}
	
	return v ;
}


extern "C" {
	static void	id3lib_ID3_def_const( )
	{
		/* ID3_TagType  */
		rb_define_const(rb_mID3Lib, "ID3TT_NONE", INT2FIX(ID3TT_NONE)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_ID3V1", INT2FIX(ID3TT_ID3V1)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_ID3V2", INT2FIX(ID3TT_ID3V2)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_LYRICS3", INT2FIX(ID3TT_LYRICS3)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_LYRICS3V2", INT2FIX(ID3TT_LYRICS3V2)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_MUSICMATCH", INT2FIX(ID3TT_MUSICMATCH)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_LYRICS", INT2FIX(ID3TT_LYRICS)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_ID3", INT2FIX(ID3TT_ID3)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_ALL", INT2FIX(ID3TT_ALL)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_PREPENDED", INT2FIX(ID3TT_PREPENDED)) ;
		rb_define_const(rb_mID3Lib, "ID3TT_APPENDED", INT2FIX(ID3TT_APPENDED)) ;
		
		/* ID3_FieldID */
		rb_define_const(rb_mID3Lib, "ID3FN_TEXTENC", INT2FIX(ID3FN_TEXTENC)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_TEXT", INT2FIX(ID3FN_TEXT)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_URL", INT2FIX(ID3FN_URL)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_DATA", INT2FIX(ID3FN_DATA)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_DESCRIPTION", INT2FIX(ID3FN_DESCRIPTION)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_OWNER", INT2FIX(ID3FN_OWNER)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_EMAIL", INT2FIX(ID3FN_EMAIL)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_RATING", INT2FIX(ID3FN_RATING)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_FILENAME", INT2FIX(ID3FN_FILENAME)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_LANGUAGE", INT2FIX(ID3FN_LANGUAGE)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_PICTURETYPE", INT2FIX(ID3FN_PICTURETYPE)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_IMAGEFORMAT", INT2FIX(ID3FN_IMAGEFORMAT)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_MIMETYPE", INT2FIX(ID3FN_MIMETYPE)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_COUNTER", INT2FIX(ID3FN_COUNTER)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_ID", INT2FIX(ID3FN_ID)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_VOLUMEADJ", INT2FIX(ID3FN_VOLUMEADJ)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_NUMBITS", INT2FIX(ID3FN_NUMBITS)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_VOLCHGRIGHT", INT2FIX(ID3FN_VOLCHGRIGHT)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_VOLCHGLEFT", INT2FIX(ID3FN_VOLCHGLEFT)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_PEAKVOLRIGHT", INT2FIX(ID3FN_PEAKVOLRIGHT)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_PEAKVOLLEFT", INT2FIX(ID3FN_PEAKVOLLEFT)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_TIMESTAMPFORMAT", INT2FIX(ID3FN_TIMESTAMPFORMAT)) ;
		rb_define_const(rb_mID3Lib, "ID3FN_CONTENTTYPE", INT2FIX(ID3FN_CONTENTTYPE)) ;
		
		/* ID3_FrameID */
		rb_define_const(rb_mID3Lib, "ID3FID_AUDIOCRYPTO", INT2FIX(ID3FID_AUDIOCRYPTO)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PICTURE", INT2FIX(ID3FID_PICTURE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_AUDIOSEEKPOINT", INT2FIX(ID3FID_AUDIOSEEKPOINT)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_COMMENT", INT2FIX(ID3FID_COMMENT)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_COMMERCIAL", INT2FIX(ID3FID_COMMERCIAL)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_CRYPTOREG", INT2FIX(ID3FID_CRYPTOREG)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_EQUALIZATION2", INT2FIX(ID3FID_EQUALIZATION2)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_EQUALIZATION", INT2FIX(ID3FID_EQUALIZATION)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_EVENTTIMING", INT2FIX(ID3FID_EVENTTIMING)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_GENERALOBJECT", INT2FIX(ID3FID_GENERALOBJECT)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_GROUPINGREG", INT2FIX(ID3FID_GROUPINGREG)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_INVOLVEDPEOPLE", INT2FIX(ID3FID_INVOLVEDPEOPLE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_LINKEDINFO", INT2FIX(ID3FID_LINKEDINFO)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_CDID", INT2FIX(ID3FID_CDID)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_MPEGLOOKUP", INT2FIX(ID3FID_MPEGLOOKUP)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_OWNERSHIP", INT2FIX(ID3FID_OWNERSHIP)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PRIVATE", INT2FIX(ID3FID_PRIVATE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PLAYCOUNTER", INT2FIX(ID3FID_PLAYCOUNTER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_POPULARIMETER", INT2FIX(ID3FID_POPULARIMETER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_POSITIONSYNC", INT2FIX(ID3FID_POSITIONSYNC)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_BUFFERSIZE", INT2FIX(ID3FID_BUFFERSIZE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_VOLUMEADJ2", INT2FIX(ID3FID_VOLUMEADJ2)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_VOLUMEADJ", INT2FIX(ID3FID_VOLUMEADJ)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_REVERB", INT2FIX(ID3FID_REVERB)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_SEEKFRAME", INT2FIX(ID3FID_SEEKFRAME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_SIGNATURE", INT2FIX(ID3FID_SIGNATURE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_SYNCEDLYRICS", INT2FIX(ID3FID_SYNCEDLYRICS)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_SYNCEDTEMPO", INT2FIX(ID3FID_SYNCEDTEMPO)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ALBUM", INT2FIX(ID3FID_ALBUM)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_BPM", INT2FIX(ID3FID_BPM)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_COMPOSER", INT2FIX(ID3FID_COMPOSER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_CONTENTTYPE", INT2FIX(ID3FID_CONTENTTYPE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_COPYRIGHT", INT2FIX(ID3FID_COPYRIGHT)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_DATE", INT2FIX(ID3FID_DATE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ENCODINGTIME", INT2FIX(ID3FID_ENCODINGTIME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PLAYLISTDELAY", INT2FIX(ID3FID_PLAYLISTDELAY)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ORIGRELEASETIME", INT2FIX(ID3FID_ORIGRELEASETIME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_RECORDINGTIME", INT2FIX(ID3FID_RECORDINGTIME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_RELEASETIME", INT2FIX(ID3FID_RELEASETIME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_TAGGINGTIME", INT2FIX(ID3FID_TAGGINGTIME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_INVOLVEDPEOPLE2", INT2FIX(ID3FID_INVOLVEDPEOPLE2)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ENCODEDBY", INT2FIX(ID3FID_ENCODEDBY)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_LYRICIST", INT2FIX(ID3FID_LYRICIST)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_FILETYPE", INT2FIX(ID3FID_FILETYPE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_TIME", INT2FIX(ID3FID_TIME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_CONTENTGROUP", INT2FIX(ID3FID_CONTENTGROUP)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_TITLE", INT2FIX(ID3FID_TITLE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_SUBTITLE", INT2FIX(ID3FID_SUBTITLE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_INITIALKEY", INT2FIX(ID3FID_INITIALKEY)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_LANGUAGE", INT2FIX(ID3FID_LANGUAGE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_SONGLEN", INT2FIX(ID3FID_SONGLEN)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_MUSICIANCREDITLIST", INT2FIX(ID3FID_MUSICIANCREDITLIST)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_MEDIATYPE", INT2FIX(ID3FID_MEDIATYPE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_MOOD", INT2FIX(ID3FID_MOOD)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ORIGALBUM", INT2FIX(ID3FID_ORIGALBUM)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ORIGFILENAME", INT2FIX(ID3FID_ORIGFILENAME)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ORIGLYRICIST", INT2FIX(ID3FID_ORIGLYRICIST)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ORIGARTIST", INT2FIX(ID3FID_ORIGARTIST)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ORIGYEAR", INT2FIX(ID3FID_ORIGYEAR)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_FILEOWNER", INT2FIX(ID3FID_FILEOWNER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_LEADARTIST", INT2FIX(ID3FID_LEADARTIST)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_BAND", INT2FIX(ID3FID_BAND)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_CONDUCTOR", INT2FIX(ID3FID_CONDUCTOR)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_MIXARTIST", INT2FIX(ID3FID_MIXARTIST)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PARTINSET", INT2FIX(ID3FID_PARTINSET)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PRODUCEDNOTICE", INT2FIX(ID3FID_PRODUCEDNOTICE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PUBLISHER", INT2FIX(ID3FID_PUBLISHER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_TRACKNUM", INT2FIX(ID3FID_TRACKNUM)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_RECORDINGDATES", INT2FIX(ID3FID_RECORDINGDATES)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_NETRADIOSTATION", INT2FIX(ID3FID_NETRADIOSTATION)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_NETRADIOOWNER", INT2FIX(ID3FID_NETRADIOOWNER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_SIZE", INT2FIX(ID3FID_SIZE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ALBUMSORTORDER", INT2FIX(ID3FID_ALBUMSORTORDER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_PERFORMERSORTORDER", INT2FIX(ID3FID_PERFORMERSORTORDER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_TITLESORTORDER", INT2FIX(ID3FID_TITLESORTORDER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ISRC", INT2FIX(ID3FID_ISRC)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_ENCODERSETTINGS", INT2FIX(ID3FID_ENCODERSETTINGS)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_USERTEXT", INT2FIX(ID3FID_USERTEXT)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_YEAR", INT2FIX(ID3FID_YEAR)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_UNIQUEFILEID", INT2FIX(ID3FID_UNIQUEFILEID)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_TERMSOFUSE", INT2FIX(ID3FID_TERMSOFUSE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_UNSYNCEDLYRICS", INT2FIX(ID3FID_UNSYNCEDLYRICS)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWCOMMERCIALINFO", INT2FIX(ID3FID_WWWCOMMERCIALINFO)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWCOPYRIGHT", INT2FIX(ID3FID_WWWCOPYRIGHT)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWAUDIOFILE", INT2FIX(ID3FID_WWWAUDIOFILE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWARTIST", INT2FIX(ID3FID_WWWARTIST)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWAUDIOSOURCE", INT2FIX(ID3FID_WWWAUDIOSOURCE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWRADIOPAGE", INT2FIX(ID3FID_WWWRADIOPAGE)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWPAYMENT", INT2FIX(ID3FID_WWWPAYMENT)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWPUBLISHER", INT2FIX(ID3FID_WWWPUBLISHER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_WWWUSER", INT2FIX(ID3FID_WWWUSER)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_METACRYPTO", INT2FIX(ID3FID_METACRYPTO)) ;
		rb_define_const(rb_mID3Lib, "ID3FID_METACOMPRESSION", INT2FIX(ID3FID_METACOMPRESSION)) ;
		
		/* ID3_TextEnc */
		rb_define_const(rb_mID3Lib, "ID3TE_ISO8859_1", INT2FIX(ID3TE_ISO8859_1)) ;
		rb_define_const(rb_mID3Lib, "ID3TE_UTF16", INT2FIX(ID3TE_UTF16)) ;
		rb_define_const(rb_mID3Lib, "ID3TE_UTF16BE", INT2FIX(ID3TE_UTF16BE)) ;
		rb_define_const(rb_mID3Lib, "ID3TE_UTF8", INT2FIX(ID3TE_UTF8)) ;
		rb_define_const(rb_mID3Lib, "ID3TE_ASCII", INT2FIX(ID3TE_ASCII)) ;
		rb_define_const(rb_mID3Lib, "ID3TE_UNICODE", INT2FIX(ID3TE_UNICODE)) ;
		
	}
	
	
	void Init_id3lib(void)
	{
		rb_mID3Lib = rb_define_module("ID3Lib") ;
		
		id3lib_ID3_def_const( ) ;
		
		rb_cID3Tag = rb_define_class_under(rb_mID3Lib, "ID3Tag", rb_cObject) ;
		rb_define_alloc_func(rb_cID3Tag, id3lib_ID3Tag_alloc) ;
		rb_define_method(rb_cID3Tag, "initialize", (VALUE (*)(...))id3lib_ID3Tag_initialize, -1) ;
		rb_define_method(rb_cID3Tag, "initialize_copy", (VALUE (*)(...))id3lib_ID3Tag_init_copy, 1) ;
		
		rb_define_method(rb_cID3Tag, "clear", (VALUE (*)(...))id3lib_ID3Tag_clear, 0) ;
		rb_define_method(rb_cID3Tag, "strip", (VALUE (*)(...))id3lib_ID3Tag_strip, -1) ;
		rb_define_method(rb_cID3Tag, "update", (VALUE (*)(...))id3lib_ID3Tag_update, -1) ;
		rb_define_method(rb_cID3Tag, "link", (VALUE (*)(...))id3lib_ID3Tag_link, -1) ;
		rb_define_method(rb_cID3Tag, "find", (VALUE (*)(...))id3lib_ID3Tag_find, -1) ;
		rb_define_method(rb_cID3Tag, "find_all", (VALUE (*)(...))id3lib_ID3Tag_find_all, -1) ;
		rb_define_method(rb_cID3Tag, "numFrames", (VALUE (*)(...))id3lib_ID3Tag_numFrames, 0) ;
		rb_define_method(rb_cID3Tag, "hasChanged?", (VALUE (*)(...))id3lib_ID3Tag_hasChanged_p, 0) ;
		rb_define_method(rb_cID3Tag, "hasV1Tag?", (VALUE (*)(...))id3lib_ID3Tag_hasV1Tag_p, 0) ;
		rb_define_method(rb_cID3Tag, "hasV2Tag?", (VALUE (*)(...))id3lib_ID3Tag_hasV2Tag_p, 0) ;
		rb_define_method(rb_cID3Tag, "setPadding", (VALUE (*)(...))id3lib_ID3Tag_setPadding, 1) ;
		rb_define_method(rb_cID3Tag, "removeFrame", (VALUE (*)(...))id3lib_ID3Tag_removeFrame, 1) ;
		rb_define_method(rb_cID3Tag, "attachFrame", (VALUE (*)(...))id3lib_ID3Tag_attachFrame, 1) ;
		rb_define_method(rb_cID3Tag, "addFrame", (VALUE (*)(...))id3lib_ID3Tag_addFrame, 1) ;
		rb_define_method(rb_cID3Tag, "<<", (VALUE (*)(...))id3lib_ID3Tag_addFrame, 1) ;
		rb_define_method(rb_cID3Tag, "each_frame", (VALUE (*)(...))id3lib_ID3Tag_each_frame, 0) ;
		
		
		rb_cID3Frame = rb_define_class_under(rb_mID3Lib, "ID3Frame", rb_cObject) ;
		rb_define_alloc_func(rb_cID3Frame, id3lib_ID3Frame_alloc) ;
		rb_define_method(rb_cID3Frame, "initialize", (VALUE (*)(...))id3lib_ID3Frame_initialize, -1) ;
		rb_define_method(rb_cID3Frame, "initialize_copy", (VALUE (*)(...))id3lib_ID3Frame_init_copy, 1) ;
		
		rb_define_method(rb_cID3Frame, "clear", (VALUE (*)(...))id3lib_ID3Frame_clear, 0) ;
		rb_define_method(rb_cID3Frame, "setID", (VALUE (*)(...))id3lib_ID3Frame_setID, 1) ;
		rb_define_method(rb_cID3Frame, "getID", (VALUE (*)(...))id3lib_ID3Frame_getID, 0) ;
		rb_define_method(rb_cID3Frame, "numFields", (VALUE (*)(...))id3lib_ID3Frame_numFields, 0) ;
		rb_define_method(rb_cID3Frame, "description", (VALUE (*)(...))id3lib_ID3Frame_getDescription, 0) ;
		rb_define_method(rb_cID3Frame, "textID", (VALUE (*)(...))id3lib_ID3Frame_getTextID, 0) ;
		rb_define_method(rb_cID3Frame, "field", (VALUE (*)(...))id3lib_ID3Frame_field, 1) ;
		rb_define_method(rb_cID3Frame, "[]", (VALUE (*)(...))id3lib_ID3Frame_field, 1) ;
		rb_define_method(rb_cID3Frame, "contains?", (VALUE (*)(...))id3lib_ID3Frame_contains_p, 1) ;
		rb_define_method(rb_cID3Frame, "getCompression", (VALUE (*)(...))id3lib_ID3Frame_getCompression, 0) ;
		rb_define_method(rb_cID3Frame, "setCompression", (VALUE (*)(...))id3lib_ID3Frame_setCompression, 1) ;
		
		
		rb_cID3Field = rb_define_class_under(rb_mID3Lib, "ID3Field", rb_cObject) ;
		rb_undef_alloc_func(rb_cID3Field) ;
		rb_undef_method(CLASS_OF(rb_cID3Field), "new") ;
		
		rb_define_method(rb_cID3Field, "clear", (VALUE (*)(...))id3lib_ID3Field_clear, 0) ;
		rb_define_method(rb_cID3Field, "get", (VALUE (*)(...))id3lib_ID3Field_get, 0) ;
		rb_define_method(rb_cID3Field, "set", (VALUE (*)(...))id3lib_ID3Field_set, 1) ;
		rb_define_method(rb_cID3Field, "isEncodable?", (VALUE (*)(...))id3lib_ID3Field_isEncodable_p, 0) ;
		rb_define_method(rb_cID3Field, "getEncoding", (VALUE (*)(...))id3lib_ID3Field_getEncoding, 0) ;
		rb_define_method(rb_cID3Field, "setEncoding", (VALUE (*)(...))id3lib_ID3Field_setEncoding, 1) ;
		
		
	}

}

