#!/usr/bin/ruby -Ke

require 'nkf'
require 'uconv'
require 'yaml'
require 'id3lib'

include ID3Lib

def set_text_tag(tag, fid, content, enc=false)
   frame = tag.find(fid)
   tag.removeFrame(frame)  if frame
   
   frame = ID3Frame.new(fid)
   if enc then
      frame[ID3FN_TEXTENC].set ID3TE_UNICODE
      frame[ID3FN_TEXT].setEncoding ID3TE_UNICODE
   else
      frame[ID3FN_TEXTENC].set ID3TE_ASCII
      frame[ID3FN_TEXT].setEncoding ID3TE_ASCII
   end
   
   frame[ID3FN_TEXT].set content
   tag.attachFrame frame
end

def set_lyrics(tag, lang, lyrics, enc=false)
   fid = ID3FID_UNSYNCEDLYRICS
   
   frame = tag.find(fid)
   tag.removeFrame(frame)  if frame
   
   frame = ID3Frame.new(fid)
   
   if enc then
      frame[ID3FN_TEXTENC].set ID3TE_UNICODE
      frame[ID3FN_TEXT].setEncoding ID3TE_UNICODE
   else
      frame[ID3FN_TEXTENC].set ID3TE_ASCII
      frame[ID3FN_TEXT].setEncoding ID3TE_ASCII
   end
   
   frame[ID3FN_TEXT].set lyrics
   
   frame[ID3FN_LANGUAGE].set lang
   
   tag.attachFrame frame
end

def set_comment(tag, lang, comment, enc=false)
   fid = ID3FID_COMMENT
   
   frame = tag.find(fid)
   tag.removeFrame(frame)  if frame
   
   frame = ID3Frame.new(fid)
   
   frame[ID3FN_LANGUAGE].set lang
   
   if enc then
      frame[ID3FN_DESCRIPTION].setEncoding ID3TE_UNICODE
      frame[ID3FN_DESCRIPTION].set Uconv.u8tou2("comment")
      
      frame[ID3FN_TEXTENC].set ID3TE_UNICODE
      frame[ID3FN_TEXT].setEncoding ID3TE_UNICODE
      frame[ID3FN_TEXT].set comment
   else
      frame[ID3FN_DESCRIPTION].setEncoding ID3TE_ASCII
      frame[ID3FN_DESCRIPTION].set "comment"
      
      frame[ID3FN_TEXTENC].set ID3TE_ASCII
      frame[ID3FN_TEXT].setEncoding ID3TE_ASCII
      frame[ID3FN_TEXT].set comment
   end
   
   tag.attachFrame frame
end

def set_image(tag, mimetype, pictype, desc, image)
   imgf = nil
   tag.find_all(ID3FID_PICTURE).each do |f|
      if f[ID3FN_PICTURETYPE].get == pictype and
            f[ID3FN_DESCRIPTION].get == desc then
         imgf = f
      end
   end
   
   tag.removeFrame imgf  if imgf
   
   imgf = ID3Frame.new(ID3FID_PICTURE)
   
   imgf[ID3FN_TEXTENC].set ID3TE_ASCII
   imgf[ID3FN_MIMETYPE].set mimetype
   imgf[ID3FN_PICTURETYPE].set pictype
   imgf[ID3FN_DESCRIPTION].set desc
   imgf[ID3FN_DATA].set image
   
   tag.attachFrame imgf
end


def open_tag(filename)
   tag = ID3Tag.new
   tag.link(filename, ID3TT_ID3V2)
   tag
end


def close_tag(tag)
   tag.update
   tag.clear
end


if ARGV.size < 1 then
   print "usage: id3tag tagfile.yaml\n"
end

begin
   f = File.open(ARGV[0])
rescue Errno::ENOENT => e
   STDERR.printf "%s: file not found.\n", ARGV[0]
end

tagdef = YAML.load(f)
f.close

tagdef.each do |discdef|
   basename   = discdef["prefix"]
   album      = discdef["album"]
   def_artist = discdef["artist"]
   def_org_artist = discdef["org_artist"]
   tracks     = discdef["tracks"]
   year       = discdef["year"]
   
   tracks.each do |trk|
      trackno = trk["trackno"]
      
      if trk.has_key?("filename") then
         filename = trk["filename"]
      else
         filename = "%s_%02d.mp3"%[basename, trackno]
      end
      
      printf "%s...\n", filename
      
      title = trk["title"]
      subtitle = trk["subtitle"]
      
      comment = trk["comment"]
      
      lyrics = trk["lyrics"]
      lyricist = trk["lyricist"]
      
      composer = trk["composer"]
      
      pict_mimetype, pict_filename = trk["cover_picture"]
      
      if trk.has_key?("artist") then
         artist = trk["artist"]
      else
         artist = def_artist
      end
      
      if trk.has_key?("org_artist") then
         org_artist = trk["org_artist"]
      else
         org_artist = def_org_artist
      end
      
      unless FileTest.file?(filename) then
         STDERR.printf "%s : file not found.\n", filename
         next
      end
      
      tag = open_tag(filename)
      set_text_tag tag, ID3FID_TRACKNUM,   trackno.to_i.to_s
      set_text_tag tag, ID3FID_ALBUM,      Uconv.u8tou2(album), true
      set_text_tag tag, ID3FID_LEADARTIST, Uconv.u8tou2(artist), true
      set_text_tag tag, ID3FID_TITLE,      Uconv.u8tou2(title), true
      
      if subtitle then
         set_text_tag tag, ID3FID_SUBTITLE, Uconv.u8tou2(subtitle), true
      end
      
      if org_artist then
         set_text_tag tag, ID3FID_ORIGARTIST, Uconv.u8tou2(org_artist), true
      end
      
      if year then
         set_text_tag tag, ID3FID_YEAR, year.to_s
      end
      
      if comment then
         set_comment tag, "jpn", Uconv.u8tou2(comment), true
      end
      
      if lyrics then
         set_lyrics tag, "jpn", Uconv.u8tou2(lyrics), true
      end
      
      if lyricist then
         set_text_tag tag, ID3FID_LYRICIST, Uconv.u8tou2(lyricist), true
      end
      
      if composer then
         set_text_tag tag, ID3FID_COMPOSER, Uconv.u8tou2(composer), true
      end
      
      if pict_filename then
         f = File.open(pict_filename, "r")
         image = f.read
         f.close
         
         set_image tag, pict_mimetype, 0x03, "Cover", image
      end
      
      close_tag(tag)
   end
end
