#!/usr/bin/ruby1.8 -Ku

Thread.abort_on_exception = true

require 'nkf'
require 'uconv'
require 'id3lib'

include ID3Lib

class MPG321
   def initialize(path)
      unless FileTest.executable?(path) then
         raise "Can't execute %s\n"%path
      end
      
      @io = IO.popen("-","a+")
      if @io.nil? then
         exec path, "-R", "-"
      end
      
      @finished = false
      @play_state = 0
      @monitor_reset = false
      @monitor_thread = Thread.start{ self.monitor }
   end
   
   def monitor
      @monitor_reset = false
      old_time = 0.0
      total_time = nil
      
      until @finished
         if @monitor_reset then
            old_time = 0.0
            total_time = nil
            @monitor_reset = false
         end
         
         ret = IO.select [@io], [], [], 1.0
         next  if ret.nil?
         
         l = @io.gets
         l.chomp!
         l.strip!
         
         case l
         when /^@S/
            @play_state = 2
            
         when /^@P (\d+)$/
            @play_state = $1.to_i
            
         when /^@F (.+) (.+) (.+) (.+)/
            current_time = $3.to_f
            remain_time = $4.to_f
            
            if total_time.nil? then
               total_time = current_time + remain_time
               tmin = (total_time/60).floor
               tsec = total_time - tmin*60.0
            end
            
            if current_time - old_time > 0.2 then
               old_time += 0.2 * ((current_time - old_time)/0.2).floor
               
               min = (current_time/60).floor
               sec = current_time - min*60.0
               STDOUT.printf "        \rtime: % 3d:%05.2f / % 3d:%05.2f", min, sec, tmin, tsec
               STDOUT.flush
            end
            
         end
         
      end
   end
   
   def reset_monitor
      @monitor_reset = true
   end
   
   def load(filename)
      @play_state = 2
      @io.printf "LOAD %s\n", filename
   end
   
   def stop
      @io.printf "STOP\n"
   end
   
   def pause
      @io.printf "PAUSE\n"
   end
   
   def quit
      @io.printf "QUIT\n"
      @finished = true
   end
   
   def play_state
      return @play_state
   end
end


def get_textframe(f)
   if f[ID3FN_TEXT].isEncodable? then
      case f[ID3FN_TEXT].getEncoding
      when ID3TE_ASCII
         return Uconv.euctou8(NKF.nkf("-e",f[ID3FN_TEXT].get))
      when ID3TE_UNICODE
         return Uconv.u2tou8(f[ID3FN_TEXT].get)
      end
   else
      return f[ID3FN_TEXT].get
   end
end

def print_id3tag(filename)
   tag = ID3Tag.new
   tag.link(filename)
   
   [
      [ID3FID_ALBUM     ,"    アルバム"],
      [ID3FID_YEAR      ,"          年"],
      [ID3FID_TITLE     ,"    タイトル"],
      [ID3FID_SUBTITLE  ,"サブタイトル"],
      [ID3FID_LEADARTIST,"アーティスト"],
      [ID3FID_COMPOSER  ,"        作曲"],
      [ID3FID_LYRICIST  ,"        作詞"]
   ].each do |fid,name|
      f = tag.find(fid)
      next  unless f
      
      printf "%s: %s\n", name, get_textframe(f)
   end
   
   fary = tag.find_all(ID3FID_COMMENT)
   fary.each do |f|
      if f[ID3FN_LANGUAGE].get == "jpn" then
         printf "コメント:\n"
         
         case f[ID3FN_TEXT].getEncoding
         when ID3TE_UNICODE
            comment = Uconv.u2tou8(f[ID3FN_TEXT].get)
         when ID3TE_ASCII
            comment = Uconv.sjistou8(f[ID3FN_TEXT].get)
         end
         
         comment.each do |l|
            printf "    %s",l
         end
         
         printf "\n"
      end
   end
   
   f = tag.find(ID3FID_UNSYNCEDLYRICS)
   if f then
      printf "歌詞:\n"
      
      case f[ID3FN_TEXT].getEncoding
      when ID3TE_ASCII
         lyrics = Uconv.sjistou8(f[ID3FN_TEXT].get)
      when ID3TE_UNICODE
         lyrics = Uconv.u2tou8(f[ID3FN_TEXT].get)
      end
      
      lyrics.each do |l|
         printf "    %s",l
      end
      
      printf "\n"
   end
   
   tag.clear
end

mpg321 = MPG321.new("/usr/bin/mpg321")

begin
   interrupted = false
   
   ARGV.each do |filename|
      unless FileTest.file?(filename) then
         STDERR.printf "%s: file not found.\n", filename
         exit
      end
      
      printf "-- \n"
      printf "playing %s\n", filename
      print_id3tag filename
      
      mpg321.load filename
      
      while mpg321.play_state != 0
         begin
            sleep 1
         rescue Interrupt => e
            interrupted = true
            break
         end
      end
      
      printf "\n"
      
      if interrupted then
         break
      end
      
      mpg321.reset_monitor
   end
   
ensure
   mpg321.quit
end

